/**
 * Copyright (c) 2022 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "DawnGLFWHost.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

// Host
DawnHost::DawnHost() {
  this->data = std::make_unique<DawnHostData>();
  this->data->window = nullptr;
}

int32_t DawnHost::init(std::weak_ptr<DawnGame> game) {
  // Init GLFW
  if(!glfwInit()) return DAWN_GLFW_INIT_RESULT_INIT_FAILED;
  
  // Create Window
  this->data->window = glfwCreateWindow(
    DAWN_GLFW_WINDOW_WIDTH_DEFAULT,
    DAWN_GLFW_WINDOW_HEIGHT_DEFAULT,
    "Dawn", NULL, NULL
  );
  if(this->data->window == nullptr) {
    glfwTerminate();
    return DAWN_GLFW_INIT_RESULT_WINDOW_CREATE_FAILED;
  }

  // Load GLAD
  glfwMakeContextCurrent(this->data->window);
  // glfwSwapInterval(0); // "Vsync" disabled if uncommented
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);

  return DAWN_HOST_INIT_RESULT_SUCCESS;
}

int32_t DawnHost::start(std::weak_ptr<DawnGame> game) {
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main Render Loop 
  time = 0.0f;
  while(true) {
    // Determine the delta.
    newTime = glfwGetTime();
    fDelta = (float_t)(newTime - time);
    time = newTime;

    // Perform update
    updateResult = this->update(game, fDelta);

    // Did the update complete successfully?
    if(updateResult == DAWN_HOST_UPDATE_RESULT_EXIT) {
      break;
    } else if(updateResult != DAWN_HOST_UPDATE_RESULT_SUCCESS) {
      return DAWN_GLFW_START_RESULT_UPDATE_FAILED;
    }
    
    // Tick the engine.
    glfwSwapBuffers(this->data->window);
    
    // Update events
    glfwPollEvents();

    // Was the window requested to close?
    if(glfwWindowShouldClose(this->data->window)) {
      break;
    }
  }

  return DAWN_HOST_START_RESULT_EXIT_SUCCESS;
}

int32_t DawnHost::update(std::weak_ptr<DawnGame> game, float_t delta) {
  if(auto g = game.lock()) {
    auto ret = g->update(delta);
    switch(ret) {
      case DAWN_GAME_UPDATE_RESULT_SUCCESS:
        return DAWN_HOST_UPDATE_RESULT_SUCCESS;
      
      case DAWN_GAME_UPDATE_RESULT_EXIT:
        return DAWN_HOST_UPDATE_RESULT_EXIT;

      default:
        return ret;
    }
  }
}

void DawnHost::unload(std::weak_ptr<DawnGame> game) {
  glfwDestroyWindow(this->data->window);
  this->data->window = nullptr;
  glfwTerminate();
}

DawnHost::~DawnHost() {

}